@@ -1,46 +1,32 @@
-// Copyright Joyent, Inc. and other Node contributors.
-//
-// Permission is hereby granted, free of charge, to any person obtaining a
-// copy of this software and associated documentation files (the
-// "Software"), to deal in the Software without restriction, including
-// without limitation the rights to use, copy, modify, merge, publish,
-// distribute, sublicense, and/or sell copies of the Software, and to permit
-// persons to whom the Software is furnished to do so, subject to the
-// following conditions:
-//
-// The above copyright notice and this permission notice shall be included
-// in all copies or substantial portions of the Software.
-//
-// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
-// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
-// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
-// USE OR OTHER DEALINGS IN THE SOFTWARE.
+'use strict';
 
-var util = require('util'),
-    path = require('path'),
-    net = require('net'),
-    vm = require('vm'),
-    repl = require('repl'),
-    inherits = util.inherits,
-    spawn = require('child_process').spawn;
+const util = require('util');
+const path = require('path');
+const net = require('net');
+const vm = require('vm');
+const Module = require('module');
+const repl = require('repl');
+const inherits = util.inherits;
+const assert = require('assert');
+const spawn = require('child_process').spawn;
+const Buffer = require('buffer').Buffer;
 
 exports.start = function(argv, stdin, stdout) {
   argv || (argv = process.argv.slice(2));
 
   if (argv.length < 1) {
     console.error('Usage: node debug script.js');
+    console.error('       node debug <host>:<port>');
+    console.error('       node debug -p <pid>');
     process.exit(1);
   }
 
   // Setup input/output streams
-  stdin = stdin || process.openStdin();
+  stdin = stdin || process.stdin;
   stdout = stdout || process.stdout;
 
-  var args = ['--debug-brk'].concat(argv),
-      interface_ = new Interface(stdin, stdout, args);
+  const args = ['--debug-brk'].concat(argv);
+  const interface_ = new Interface(stdin, stdout, args);
 
   stdin.resume();
 
@@ -59,7 +45,7 @@ exports.port = 5858;
 
 //
 // Parser/Serializer for V8 debugger protocol
-// http://code.google.com/p/v8/wiki/DebuggerProtocol
+// https://github.com/v8/v8/wiki/Debugging-Protocol
 //
 // Usage:
 //    p = new Protocol();
@@ -116,7 +102,7 @@ Protocol.prototype.execute = function(d) {
       if (len - this.bodyStartByteIndex < this.contentLength) {
         break;
       }
-      // pass thru
+      // falls through
     case 'body':
       var resRawByteLength = Buffer.byteLength(res.raw, 'utf8');
 
@@ -140,7 +126,6 @@ Protocol.prototype.execute = function(d) {
 
     default:
       throw new Error('Unknown state');
-      break;
   }
 };
 
@@ -154,7 +139,7 @@ Protocol.prototype.serialize = function(req) {
 };
 
 
-var NO_FRAME = -1;
+const NO_FRAME = -1;
 
 function Client() {
   net.Stream.call(this);
@@ -164,7 +149,6 @@ function Client() {
 
   this.currentFrame = NO_FRAME;
   this.currentSourceLine = -1;
-  this.currentSource = null;
   this.handles = {};
   this.scripts = {};
   this.breakpoints = [];
@@ -175,26 +159,27 @@ function Client() {
     protocol.execute(d);
   });
 
-  protocol.onResponse = this._onResponse.bind(this);
+  protocol.onResponse = (res) => this._onResponse(res);
 }
 inherits(Client, net.Stream);
 exports.Client = Client;
 
 
 Client.prototype._addHandle = function(desc) {
-  if (typeof desc != 'object' || typeof desc.handle != 'number') {
+  if (desc === null || typeof desc !== 'object' ||
+      typeof desc.handle !== 'number') {
     return;
   }
 
   this.handles[desc.handle] = desc;
 
-  if (desc.type == 'script') {
+  if (desc.type === 'script') {
     this._addScript(desc);
   }
 };
 
 
-var natives = process.binding('natives');
+const natives = process.binding('natives');
 
 
 Client.prototype._addScript = function(desc) {
@@ -212,8 +197,8 @@ Client.prototype._removeScript = function(desc) {
 
 
 Client.prototype._onResponse = function(res) {
-  var cb,
-      index = -1;
+  var cb;
+  var index = -1;
 
   this._reqCallbacks.some(function(fn, i) {
     if (fn.request_seq == res.body.request_seq) {
@@ -249,6 +234,10 @@ Client.prototype._onResponse = function(res) {
     this._removeScript(res.body.body.script);
     handled = true;
 
+  } else if (res.body && res.body.event === 'compileError') {
+    // This event is not used anywhere right now, perhaps somewhere in the
+    // future?
+    handled = true;
   }
 
   if (cb) {
@@ -273,7 +262,7 @@ Client.prototype.req = function(req, cb) {
 
 Client.prototype.reqVersion = function(cb) {
   cb = cb || function() {};
-  this.req({ command: 'version' } , function(err, body, res) {
+  this.req({ command: 'version' }, function(err, body, res) {
     if (err) return cb(err);
     cb(null, res.body.body.V8Version, res.body.running);
   });
@@ -296,7 +285,7 @@ Client.prototype.reqLookup = function(refs, cb) {
   this.req(req, function(err, res) {
     if (err) return cb(err);
     for (var ref in res) {
-      if (typeof res[ref] == 'object') {
+      if (res[ref] !== null && typeof res[ref] === 'object') {
         self._addHandle(res[ref]);
       }
     }
@@ -306,11 +295,11 @@ Client.prototype.reqLookup = function(refs, cb) {
 };
 
 Client.prototype.reqScopes = function(cb) {
-  var self = this,
-      req = {
-        command: 'scopes',
-        arguments: {}
-      };
+  const self = this;
+  const req = {
+    command: 'scopes',
+    arguments: {}
+  };
 
   cb = cb || function() {};
   this.req(req, function(err, res) {
@@ -366,7 +355,7 @@ Client.prototype.reqEval = function(expression, cb) {
 };
 
 
-// Finds the first scope in the array in which the epxression evals.
+// Finds the first scope in the array in which the expression evals.
 Client.prototype._reqFramesEval = function(expression, evalFrames, cb) {
   if (evalFrames.length == 0) {
     // Just eval in global scope.
@@ -409,7 +398,7 @@ Client.prototype.reqFrameEval = function(expression, frame, cb) {
 // reqBacktrace(cb)
 // TODO: from, to, bottom
 Client.prototype.reqBacktrace = function(cb) {
-  this.req({ command: 'backtrace', arguments: { inlineRefs: true } } , cb);
+  this.req({ command: 'backtrace', arguments: { inlineRefs: true } }, cb);
 };
 
 
@@ -443,7 +432,7 @@ Client.prototype.reqScripts = function(cb) {
   var self = this;
   cb = cb || function() {};
 
-  this.req({ command: 'scripts' } , function(err, res) {
+  this.req({ command: 'scripts' }, function(err, res) {
     if (err) return cb(err);
 
     for (var i = 0; i < res.length; i++) {
@@ -473,7 +462,7 @@ Client.prototype.setBreakpoint = function(req, cb) {
 };
 
 Client.prototype.clearBreakpoint = function(req, cb) {
-  var req = {
+  req = {
     command: 'clearbreakpoint',
     arguments: req
   };
@@ -536,8 +525,8 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
         return;
       }
 
-      var mirror,
-          waiting = 1;
+      var mirror;
+      var waiting = 1;
 
       if (handle.className == 'Array') {
         mirror = [];
@@ -558,9 +547,7 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
           mirrorValue = '[?]';
         }
 
-
-        if (Array.isArray(mirror) &&
-            typeof prop.name != 'number') {
+        if (Array.isArray(mirror) && typeof prop.name !== 'number') {
           // Skip the 'length' property.
           return;
         }
@@ -580,13 +567,13 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
 
       waitForOthers();
       function waitForOthers() {
-        if (--waiting === 0 && cb) {
+        if (--waiting === 0) {
           keyValues.forEach(function(kv) {
             mirror[kv.name] = kv.value;
           });
           cb(null, mirror);
         }
-      };
+      }
     });
     return;
   } else if (handle.type === 'function') {
@@ -600,9 +587,7 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
   } else {
     val = handle;
   }
-  process.nextTick(function() {
-    cb(null, val);
-  });
+  process.nextTick(cb, null, val);
 };
 
 
@@ -656,11 +641,7 @@ Client.prototype.fullTrace = function(cb) {
 };
 
 
-
-
-
-
-var commands = [
+const commands = [
   [
     'run (r)',
     'cont (c)',
@@ -695,8 +676,8 @@ var helpMessage = 'Commands: ' + commands.map(function(group) {
 function SourceUnderline(sourceText, position, repl) {
   if (!sourceText) return '';
 
-  var head = sourceText.slice(0, position),
-      tail = sourceText.slice(position);
+  const head = sourceText.slice(0, position);
+  var tail = sourceText.slice(position);
 
   // Colourize char if stdout supports colours
   if (repl.useColors) {
@@ -716,8 +697,8 @@ function SourceInfo(body) {
 
   if (body.script) {
     if (body.script.name) {
-      var name = body.script.name,
-          dir = path.resolve() + '/';
+      var name = body.script.name;
+      const dir = path.resolve() + '/';
 
       // Change path to relative, if possible
       if (name.indexOf(dir) === 0) {
@@ -740,8 +721,7 @@ function SourceInfo(body) {
 // This class is the repl-enabled debugger interface which is invoked on
 // "node debug"
 function Interface(stdin, stdout, args) {
-  var self = this,
-      child;
+  var self = this;
 
   this.stdin = stdin;
   this.stdout = stdout;
@@ -753,16 +733,26 @@ function Interface(stdin, stdout, args) {
     prompt: 'debug> ',
     input: this.stdin,
     output: this.stdout,
-    eval: this.controlEval.bind(this),
+    eval: (code, ctx, file, cb) => this.controlEval(code, ctx, file, cb),
     useGlobal: false,
     ignoreUndefined: true
   };
   if (parseInt(process.env['NODE_NO_READLINE'], 10)) {
     opts.terminal = false;
+  } else if (parseInt(process.env['NODE_FORCE_READLINE'], 10)) {
+    opts.terminal = true;
+
+    // Emulate Ctrl+C if we're emulating terminal
+    if (!this.stdout.isTTY) {
+      process.on('SIGINT', function() {
+        self.repl.rli.emit('SIGINT');
+      });
+    }
   }
   if (parseInt(process.env['NODE_DISABLE_COLORS'], 10)) {
     opts.useColors = false;
   }
+
   this.repl = repl.start(opts);
 
   // Do not print useless warning
@@ -773,27 +763,27 @@ function Interface(stdin, stdout, args) {
     process.exit(0);
   });
 
-  process.on('exit', function() {
-    self.killChild();
-  });
+  // Handle all possible exits
+  process.on('exit', () => this.killChild());
+  process.once('SIGTERM', process.exit.bind(process, 0));
+  process.once('SIGHUP', process.exit.bind(process, 0));
 
-
-  var proto = Interface.prototype,
-      ignored = ['pause', 'resume', 'exitRepl', 'handleBreak',
-                 'requireConnection', 'killChild', 'trySpawn',
-                 'controlEval', 'debugEval', 'print', 'childPrint',
-                 'clearline'],
-      shortcut = {
-        'run': 'r',
-        'cont': 'c',
-        'next': 'n',
-        'step': 's',
-        'out': 'o',
-        'backtrace': 'bt',
-        'setBreakpoint': 'sb',
-        'clearBreakpoint': 'cb',
-        'pause_': 'pause'
-      };
+  var proto = Interface.prototype;
+  const ignored = ['pause', 'resume', 'exitRepl', 'handleBreak',
+                   'requireConnection', 'killChild', 'trySpawn',
+                   'controlEval', 'debugEval', 'print', 'childPrint',
+                   'clearline'];
+  const shortcut = {
+    'run': 'r',
+    'cont': 'c',
+    'next': 'n',
+    'step': 's',
+    'out': 'o',
+    'backtrace': 'bt',
+    'setBreakpoint': 'sb',
+    'clearBreakpoint': 'cb',
+    'pause_': 'pause'
+  };
 
   function defineProperty(key, protoKey) {
     // Check arity
@@ -808,7 +798,7 @@ function Interface(stdin, stdout, args) {
     } else {
       self.repl.context[key] = fn;
     }
-  };
+  }
 
   // Copy all prototype methods in repl context
   // Setup them as getters if possible
@@ -848,13 +838,14 @@ function Interface(stdin, stdout, args) {
 
 
 Interface.prototype.pause = function() {
-  if (this.killed || this.paused++ > 0) return false;
+  if (this.killed || this.paused++ > 0) return this;
   this.repl.rli.pause();
   this.stdin.pause();
+  return this;
 };
 
 Interface.prototype.resume = function(silent) {
-  if (this.killed || this.paused === 0 || --this.paused !== 0) return false;
+  if (this.killed || this.paused === 0 || --this.paused !== 0) return this;
   this.repl.rli.resume();
   if (silent !== true) {
     this.repl.displayPrompt();
@@ -865,6 +856,7 @@ Interface.prototype.resume = function(silent) {
     this.waiting();
     this.waiting = null;
   }
+  return this;
 };
 
 
@@ -952,8 +944,8 @@ Interface.prototype.controlEval = function(code, context, filename, callback) {
   try {
     // Repeat last command if empty line are going to be evaluated
     if (this.repl.rli.history && this.repl.rli.history.length > 0) {
-      if (code === '(undefined\n)') {
-        code = '(' + this.repl.rli.history[0] + '\n)';
+      if (code === '\n') {
+        code = this.repl.rli.history[0] + '\n';
       }
     }
 
@@ -977,8 +969,8 @@ Interface.prototype.controlEval = function(code, context, filename, callback) {
 Interface.prototype.debugEval = function(code, context, filename, callback) {
   if (!this.requireConnection()) return;
 
-  var self = this,
-      client = this.client;
+  const self = this;
+  const client = this.client;
 
   // Repl asked for scope variables
   if (code === '.scope') {
@@ -1009,33 +1001,14 @@ Interface.prototype.debugEval = function(code, context, filename, callback) {
 
 // Utils
 
-// Returns number of digits (+1)
-function intChars(n) {
-  // TODO dumb:
-  if (n < 50) {
-    return 3;
-  } else if (n < 950) {
-    return 4;
-  } else if (n < 9950) {
-    return 5;
-  } else {
-    return 6;
-  }
-}
-
 // Adds spaces and prefix to number
-function leftPad(n, prefix) {
-  var s = n.toString(),
-      nchars = intChars(n),
-      nspaces = nchars - s.length - 1;
+// maxN is a maximum number we should have space for
+function leftPad(n, prefix, maxN) {
+  const s = n.toString();
+  const nchars = Math.max(2, String(maxN).length) + 1;
+  const nspaces = nchars - s.length - 1;
 
-  prefix || (prefix = ' ');
-
-  for (var i = 0; i < nspaces; i++) {
-    prefix += ' ';
-  }
-
-  return prefix + s;
+  return prefix + ' '.repeat(nspaces) + s;
 }
 
 
@@ -1101,10 +1074,10 @@ Interface.prototype.list = function(delta) {
 
   delta || (delta = 5);
 
-  var self = this,
-      client = this.client,
-      from = client.currentSourceLine - delta + 1,
-      to = client.currentSourceLine + delta + 1;
+  const self = this;
+  const client = this.client;
+  const from = client.currentSourceLine - delta + 1;
+  const to = client.currentSourceLine + delta + 1;
 
   self.pause();
   client.reqSource(from, to, function(err, res) {
@@ -1119,16 +1092,17 @@ Interface.prototype.list = function(delta) {
       var lineno = res.fromLine + i + 1;
       if (lineno < from || lineno > to) continue;
 
-      var current = lineno == 1 + client.currentSourceLine,
-          breakpoint = client.breakpoints.some(function(bp) {
-            return bp.script === client.currentScript &&
-                   bp.line == lineno;
-          });
+      const current = lineno == 1 + client.currentSourceLine;
+      const breakpoint = client.breakpoints.some(function(bp) {
+        return (bp.scriptReq === client.currentScript ||
+                bp.script === client.currentScript) &&
+                bp.line == lineno;
+      });
 
       if (lineno == 1) {
         // The first line needs to have the module wrapper filtered out of
         // it.
-        var wrapper = require('module').wrapper[0];
+        var wrapper = Module.wrapper[0];
         lines[i] = lines[i].slice(wrapper.length);
 
         client.currentSourceColumn -= wrapper.length;
@@ -1144,7 +1118,14 @@ Interface.prototype.list = function(delta) {
         line = lines[i];
       }
 
-      self.print(leftPad(lineno, breakpoint && '*') + ' ' + line);
+      var prefixChar = ' ';
+      if (current) {
+        prefixChar = '>';
+      } else if (breakpoint) {
+        prefixChar = '*';
+      }
+
+      self.print(leftPad(lineno, prefixChar, to) + ' ' + line);
     }
     self.resume();
   });
@@ -1154,8 +1135,8 @@ Interface.prototype.list = function(delta) {
 Interface.prototype.backtrace = function() {
   if (!this.requireConnection()) return;
 
-  var self = this,
-      client = this.client;
+  const self = this;
+  const client = this.client;
 
   self.pause();
   client.fullTrace(function(err, bt) {
@@ -1168,8 +1149,8 @@ Interface.prototype.backtrace = function() {
     if (bt.totalFrames == 0) {
       self.print('(empty stack)');
     } else {
-      var trace = [],
-          firstFrameNative = bt.frames[0].script.isNative;
+      const trace = [];
+      const firstFrameNative = bt.frames[0].script.isNative;
 
       for (var i = 0; i < bt.frames.length; i++) {
         var frame = bt.frames[i];
@@ -1198,14 +1179,14 @@ Interface.prototype.backtrace = function() {
 Interface.prototype.scripts = function() {
   if (!this.requireConnection()) return;
 
-  var client = this.client,
-      displayNatives = arguments[0] || false,
-      scripts = [];
+  const client = this.client;
+  const displayNatives = arguments[0] || false;
+  const scripts = [];
 
   this.pause();
   for (var id in client.scripts) {
     var script = client.scripts[id];
-    if (typeof script == 'object' && script.name) {
+    if (script !== null && typeof script === 'object' && script.name) {
       if (displayNatives ||
           script.name == client.currentScript ||
           !script.isNative) {
@@ -1280,11 +1261,11 @@ Interface.prototype.unwatch = function(expr) {
 
 // List watchers
 Interface.prototype.watchers = function() {
-  var self = this,
-      verbose = arguments[0] || false,
-      callback = arguments[1] || function() {},
-      waiting = this._watchers.length,
-      values = [];
+  var self = this;
+  var verbose = arguments[0] || false;
+  var callback = arguments[1] || function() {};
+  var waiting = this._watchers.length;
+  var values = [];
 
   this.pause();
 
@@ -1306,7 +1287,8 @@ Interface.prototype.watchers = function() {
       if (verbose) self.print('Watchers:');
 
       self._watchers.forEach(function(watcher, i) {
-        self.print(leftPad(i, ' ') + ': ' + watcher + ' = ' +
+        self.print(leftPad(i, ' ', self._watchers.length - 1) +
+                   ': ' + watcher + ' = ' +
                    JSON.stringify(values[i]));
       });
 
@@ -1337,9 +1319,9 @@ Interface.prototype.setBreakpoint = function(script, line,
                                              condition, silent) {
   if (!this.requireConnection()) return;
 
-  var self = this,
-      scriptId,
-      ambiguous;
+  const self = this;
+  var scriptId;
+  var ambiguous;
 
   // setBreakpoint() should insert breakpoint on current line
   if (script === undefined) {
@@ -1353,9 +1335,16 @@ Interface.prototype.setBreakpoint = function(script, line,
     script = this.client.currentScript;
   }
 
-  if (/\(\)$/.test(script)) {
+  if (script === undefined) {
+    this.print('Cannot determine the current script, ' +
+        'make sure the debugged process is paused.');
+    return;
+  }
+
+  let req;
+  if (script.endsWith('()')) {
     // setBreakpoint('functionname()');
-    var req = {
+    req = {
       type: 'function',
       target: script.replace(/\(\)$/, ''),
       condition: condition
@@ -1364,28 +1353,41 @@ Interface.prototype.setBreakpoint = function(script, line,
     // setBreakpoint('scriptname')
     if (script != +script && !this.client.scripts[script]) {
       var scripts = this.client.scripts;
-      Object.keys(scripts).forEach(function(id) {
-        if (scripts[id] && scripts[id].name.indexOf(script) !== -1) {
+      for (var id in scripts) {
+        if (scripts[id] &&
+            scripts[id].name &&
+            scripts[id].name.indexOf(script) !== -1) {
           if (scriptId) {
             ambiguous = true;
           }
           scriptId = id;
         }
-      });
+      }
     } else {
       scriptId = script;
     }
 
-    if (!scriptId) return this.error('Script : ' + script + ' not found');
     if (ambiguous) return this.error('Script name is ambiguous');
     if (line <= 0) return this.error('Line should be a positive value');
 
-    var req = {
-      type: 'scriptId',
-      target: scriptId,
-      line: line - 1,
-      condition: condition
-    };
+    if (scriptId) {
+      req = {
+        type: 'scriptId',
+        target: scriptId,
+        line: line - 1,
+        condition: condition
+      };
+    } else {
+      this.print('Warning: script \'' + script + '\' was not loaded yet.');
+      var escapedPath = script.replace(/([/\\.?*()^${}|[\]])/g, '\\$1');
+      var scriptPathRegex = '^(.*[\\/\\\\])?' + escapedPath + '$';
+      req = {
+        type: 'scriptRegExp',
+        target: scriptPathRegex,
+        line: line - 1,
+        condition: condition
+      };
+    }
   }
 
   self.pause();
@@ -1402,20 +1404,18 @@ Interface.prototype.setBreakpoint = function(script, line,
       // Try load scriptId and line from response
       if (!scriptId) {
         scriptId = res.script_id;
-        line = res.line;
-      }
-
-      // If we finally have one - remember this breakpoint
-      if (scriptId) {
-        self.client.breakpoints.push({
-          id: res.breakpoint,
-          scriptId: scriptId,
-          script: (self.client.scripts[scriptId] || {}).name,
-          line: line,
-          condition: condition
-        });
+        line = res.line + 1;
       }
 
+      // Remember this breakpoint even if scriptId is not resolved yet
+      self.client.breakpoints.push({
+        id: res.breakpoint,
+        scriptId: scriptId,
+        script: (self.client.scripts[scriptId] || {}).name,
+        line: line,
+        condition: condition,
+        scriptReq: script
+      });
     }
     self.resume();
   });
@@ -1425,15 +1425,19 @@ Interface.prototype.setBreakpoint = function(script, line,
 Interface.prototype.clearBreakpoint = function(script, line) {
   if (!this.requireConnection()) return;
 
-  var ambiguous,
-      breakpoint,
-      index;
+  var ambiguous;
+  var breakpoint;
+  var scriptId;
+  var index;
 
   this.client.breakpoints.some(function(bp, i) {
-    if (bp.scriptId === script || bp.script.indexOf(script) !== -1) {
+    if (bp.scriptId === script ||
+        bp.scriptReq === script ||
+        (bp.script && bp.script.indexOf(script) !== -1)) {
       if (index !== undefined) {
         ambiguous = true;
       }
+      scriptId = script;
       if (bp.line === line) {
         index = i;
         breakpoint = bp.id;
@@ -1442,16 +1446,32 @@ Interface.prototype.clearBreakpoint = function(script, line) {
     }
   });
 
-  if (ambiguous) return this.error('Script name is ambiguous');
-
-  if (breakpoint === undefined) {
-    return this.error('Script : ' + script + ' not found');
+  if (!scriptId && !this.client.scripts[script]) {
+    var scripts = this.client.scripts;
+    for (var id in scripts) {
+      if (scripts[id] &&
+          scripts[id].name &&
+          scripts[id].name.indexOf(script) !== -1) {
+        if (scriptId) {
+          ambiguous = true;
+        }
+        scriptId = id;
+      }
+    }
   }
 
-  var self = this,
-      req = {
-        breakpoint: breakpoint
-      };
+  if (ambiguous) return this.error('Script name is ambiguous');
+
+  if (scriptId === undefined) {
+    return this.error('Script ' + script + ' not found');
+  }
+
+  if (breakpoint === undefined) {
+    return this.error('Breakpoint not found on line ' + line);
+  }
+
+  var self = this;
+  const req = {breakpoint};
 
   self.pause();
   self.client.clearBreakpoint(req, function(err, res) {
@@ -1487,8 +1507,8 @@ Interface.prototype.breakpoints = function() {
 Interface.prototype.pause_ = function() {
   if (!this.requireConnection()) return;
 
-  var self = this,
-      cmd = 'process._debugPause();';
+  const self = this;
+  const cmd = 'process._debugPause();';
 
   this.pause();
   this.client.reqFrameEval(cmd, NO_FRAME, function(err, res) {
@@ -1520,8 +1540,7 @@ Interface.prototype.repl = function() {
   var listeners = this.repl.rli.listeners('SIGINT').slice(0);
   this.repl.rli.removeAllListeners('SIGINT');
 
-  // Exit debug repl on Ctrl + C
-  this.repl.rli.once('SIGINT', function() {
+  function exitDebugRepl() {
     // Restore all listeners
     process.nextTick(function() {
       listeners.forEach(function(listener) {
@@ -1531,17 +1550,26 @@ Interface.prototype.repl = function() {
 
     // Exit debug repl
     self.exitRepl();
-  });
+
+    self.repl.rli.removeListener('SIGINT', exitDebugRepl);
+    self.repl.removeListener('exit', exitDebugRepl);
+  }
+
+  // Exit debug repl on SIGINT
+  this.repl.rli.on('SIGINT', exitDebugRepl);
+
+  // Exit debug repl on repl exit
+  this.repl.on('exit', exitDebugRepl);
 
   // Set new
-  this.repl.eval = this.debugEval.bind(this);
+  this.repl.eval = (code, ctx, file, cb) =>
+    this.debugEval(code, ctx, file, cb);
   this.repl.context = {};
 
   // Swap history
   this.history.control = this.repl.rli.history;
   this.repl.rli.history = this.history.debug;
 
-  this.repl.prompt = '> ';
   this.repl.rli.setPrompt('> ');
   this.repl.displayPrompt();
 };
@@ -1550,14 +1578,14 @@ Interface.prototype.repl = function() {
 // Exit debug repl
 Interface.prototype.exitRepl = function() {
   // Restore eval
-  this.repl.eval = this.controlEval.bind(this);
+  this.repl.eval = (code, ctx, file, cb) =>
+    this.controlEval(code, ctx, file, cb);
 
   // Swap history
   this.history.debug = this.repl.rli.history;
   this.repl.rli.history = this.history.control;
 
   this.repl.context = this.context;
-  this.repl.prompt = 'debug> ';
   this.repl.rli.setPrompt('debug> ');
   this.repl.displayPrompt();
 };
@@ -1589,65 +1617,71 @@ Interface.prototype.killChild = function() {
 
 // Spawns child process (and restores breakpoints)
 Interface.prototype.trySpawn = function(cb) {
-  var self = this,
-      breakpoints = this.breakpoints || [],
-      port = exports.port,
-      host = 'localhost';
+  const self = this;
+  const breakpoints = this.breakpoints || [];
+  var port = exports.port;
+  var host = '127.0.0.1';
+  var childArgs = this.args;
 
   this.killChild();
+  assert(!this.child);
 
+  var isRemote = false;
   if (this.args.length === 2) {
-    var match = this.args[1].match(/^([^:]+):(\d+)$/);
+    const match = this.args[1].match(/^([^:]+):(\d+)$/);
 
     if (match) {
       // Connecting to remote debugger
       // `node debug localhost:5858`
       host = match[1];
       port = parseInt(match[2], 10);
-      this.child = {
-        kill: function() {
-          // TODO Do we really need to handle it?
-        }
-      };
+      isRemote = true;
     }
   } else if (this.args.length === 3) {
     // `node debug -p pid`
     if (this.args[1] === '-p' && /^\d+$/.test(this.args[2])) {
-      this.child = {
-        kill: function() {
-          // TODO Do we really need to handle it?
+      const pid = parseInt(this.args[2], 10);
+      try {
+        process._debugProcess(pid);
+      } catch (e) {
+        if (e.code === 'ESRCH') {
+          console.error(`Target process: ${pid} doesn't exist.`);
+          process.exit(1);
         }
-      };
-      process._debugProcess(parseInt(this.args[2], 10));
+        throw e;
+      }
+      isRemote = true;
     } else {
-      var match = this.args[1].match(/^--port=(\d+)$/);
+      const match = this.args[1].match(/^--port=(\d+)$/);
       if (match) {
         // Start debugger on custom port
         // `node debug --port=5858 app.js`
         port = parseInt(match[1], 10);
-        this.args.splice(0, 2, '--debug-brk=' + port);
+        childArgs = ['--debug-brk=' + port].concat(this.args.slice(2));
       }
     }
   }
 
-  if (!this.child) {
-    this.child = spawn(process.execPath, this.args);
+  if (!isRemote) {
+    // pipe stream into debugger
+    this.child = spawn(process.execPath, childArgs);
 
-    this.child.stdout.on('data', this.childPrint.bind(this));
-    this.child.stderr.on('data', this.childPrint.bind(this));
+    this.child.stdout.on('data', (text) => this.childPrint(text));
+    this.child.stderr.on('data', (text) => this.childPrint(text));
   }
 
   this.pause();
 
-  var client = self.client = new Client(),
-      connectionAttempts = 0;
+  const client = self.client = new Client();
+  var connectionAttempts = 0;
 
   client.once('ready', function() {
     self.stdout.write(' ok\n');
 
     // Restore breakpoints
     breakpoints.forEach(function(bp) {
-      self.setBreakpoint(bp.scriptId, bp.line, bp.condition, true);
+      self.print('Restoring breakpoint ' + bp.scriptReq + ':' + bp.line);
+      self.setBreakpoint(bp.scriptReq, bp.line, bp.condition, true);
     });
 
     client.on('close', function() {
@@ -1678,9 +1712,10 @@ Interface.prototype.trySpawn = function(cb) {
 
   client.on('error', connectError);
   function connectError() {
-    // If it's failed to connect 4 times then don't catch the next error
+    // If it's failed to connect 10 times then print failed message
     if (connectionAttempts >= 10) {
-      client.removeListener('error', connectError);
+      console.error(' failed, please retry');
+      process.exit(1);
     }
     setTimeout(attemptConnect, 500);
   }
@@ -1691,8 +1726,12 @@ Interface.prototype.trySpawn = function(cb) {
     client.connect(port, host);
   }
 
-  setTimeout(function() {
-    self.print('connecting..', true);
+  self.print('connecting to ' + host + ':' + port + ' ..', true);
+  if (isRemote) {
     attemptConnect();
-  }, 50);
+  } else {
+    this.child.stderr.once('data', function() {
+      setImmediate(attemptConnect);
+    });
+  }
 };
