@@ -1,28 +1,7 @@
-// Copyright Joyent, Inc. and other Node contributors.
-//
-// Permission is hereby granted, free of charge, to any person obtaining a
-// copy of this software and associated documentation files (the
-// "Software"), to deal in the Software without restriction, including
-// without limitation the rights to use, copy, modify, merge, publish,
-// distribute, sublicense, and/or sell copies of the Software, and to permit
-// persons to whom the Software is furnished to do so, subject to the
-// following conditions:
-//
-// The above copyright notice and this permission notice shall be included
-// in all copies or substantial portions of the Software.
-//
-// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
-// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
-// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
-// USE OR OTHER DEALINGS IN THE SOFTWARE.
-
 /* A repl library that you can include in your own code to get a runtime
  * interface to your program.
  *
- *   var repl = require("/repl.js");
+ *   var repl = require("repl");
  *   // start repl on stdin
  *   repl.start("prompt> ");
  *
@@ -40,14 +19,38 @@
  *   repl.start("node > ").context.foo = "stdin is fun";
  */
 
-var util = require('util');
-var inherits = require('util').inherits;
-var Stream = require('stream');
-var vm = require('vm');
-var path = require('path');
-var fs = require('fs');
-var rl = require('readline');
-var EventEmitter = require('events').EventEmitter;
+'use strict';
+
+const internalModule = require('internal/module');
+const internalUtil = require('internal/util');
+const util = require('util');
+const inherits = util.inherits;
+const Stream = require('stream');
+const vm = require('vm');
+const path = require('path');
+const fs = require('fs');
+const rl = require('readline');
+const Console = require('console').Console;
+const Module = require('module');
+const domain = require('domain');
+const debug = util.debuglog('repl');
+
+const parentModule = module;
+const replMap = new WeakMap();
+
+try {
+  // hack for require.resolve("./relative") to work properly.
+  module.filename = path.resolve('repl');
+} catch (e) {
+  // path.resolve('repl') fails when the current working directory has been
+  // deleted.  Fall back to the directory name of the (absolute) executable
+  // path.  It's not really correct but what are the alternatives?
+  const dirname = path.dirname(process.execPath);
+  module.filename = path.resolve(dirname, 'repl');
+}
+
+// hack for repl require to work properly with node_modules folders
+module.paths = require('module')._nodeModulePaths(module.filename);
 
 // If obj.hasOwnProperty has been overridden, then calling
 // obj.hasOwnProperty(prop) will break.
@@ -57,33 +60,129 @@ function hasOwnProperty(obj, prop) {
 }
 
 
-var context;
-
-// hack for require.resolve("./relative") to work properly.
-module.filename = path.resolve('repl');
-
-// hack for repl require to work properly with node_modules folders
-module.paths = require('module')._nodeModulePaths(module.filename);
-
 // Can overridden with custom print functions, such as `probe` or `eyes.js`.
 // This is the default "writer" value if none is passed in the REPL options.
 exports.writer = util.inspect;
 
 exports._builtinLibs = ['assert', 'buffer', 'child_process', 'cluster',
-  'crypto', 'dgram', 'dns', 'events', 'fs', 'http', 'https', 'net',
-  'os', 'path', 'punycode', 'querystring', 'readline', 'repl',
-  'string_decoder', 'tls', 'tty', 'url', 'util', 'vm', 'zlib'];
+  'crypto', 'dgram', 'dns', 'domain', 'events', 'fs', 'http', 'https', 'net',
+  'os', 'path', 'punycode', 'querystring', 'readline', 'stream',
+  'string_decoder', 'tls', 'tty', 'url', 'util', 'v8', 'vm', 'zlib'];
 
 
-function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
-  if (!(this instanceof REPLServer)) {
-    return new REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined);
+const BLOCK_SCOPED_ERROR = 'Block-scoped declarations (let, ' +
+    'const, function, class) not yet supported outside strict mode';
+
+
+class LineParser {
+
+  constructor() {
+    this.reset();
   }
 
-  EventEmitter.call(this);
+  reset() {
+    this._literal = null;
+    this.shouldFail = false;
+    this.blockComment = false;
+    this.regExpLiteral = false;
+  }
 
-  var options, input, output;
-  if (typeof prompt == 'object') {
+  parseLine(line) {
+    var previous = null;
+    this.shouldFail = false;
+    const wasWithinStrLiteral = this._literal !== null;
+
+    for (const current of line) {
+      if (previous === '\\') {
+        // valid escaping, skip processing. previous doesn't matter anymore
+        previous = null;
+        continue;
+      }
+
+      if (!this._literal) {
+        if (this.regExpLiteral && current === '/') {
+          this.regExpLiteral = false;
+          previous = null;
+          continue;
+        }
+        if (previous === '*' && current === '/') {
+          if (this.blockComment) {
+            this.blockComment = false;
+            previous = null;
+            continue;
+          } else {
+            this.shouldFail = true;
+            break;
+          }
+        }
+
+        // ignore rest of the line if `current` and `previous` are `/`s
+        if (previous === current && previous === '/' && !this.blockComment) {
+          break;
+        }
+
+        if (previous === '/') {
+          if (current === '*') {
+            this.blockComment = true;
+          } else {
+            this.regExpLiteral = true;
+          }
+          previous = null;
+        }
+      }
+
+      if (this.blockComment || this.regExpLiteral) continue;
+
+      if (current === this._literal) {
+        this._literal = null;
+      } else if (current === '\'' || current === '"') {
+        this._literal = this._literal || current;
+      }
+
+      previous = current;
+    }
+
+    const isWithinStrLiteral = this._literal !== null;
+
+    if (!wasWithinStrLiteral && !isWithinStrLiteral) {
+      // Current line has nothing to do with String literals, trim both ends
+      line = line.trim();
+    } else if (wasWithinStrLiteral && !isWithinStrLiteral) {
+      // was part of a string literal, but it is over now, trim only the end
+      line = line.trimRight();
+    } else if (isWithinStrLiteral && !wasWithinStrLiteral) {
+      // was not part of a string literal, but it is now, trim only the start
+      line = line.trimLeft();
+    }
+
+    const lastChar = line.charAt(line.length - 1);
+
+    this.shouldFail = this.shouldFail ||
+                      ((!this._literal && lastChar === '\\') ||
+                      (this._literal   && lastChar !== '\\'));
+
+    return line;
+  }
+}
+
+
+function REPLServer(prompt,
+                    stream,
+                    eval_,
+                    useGlobal,
+                    ignoreUndefined,
+                    replMode) {
+  if (!(this instanceof REPLServer)) {
+    return new REPLServer(prompt,
+                          stream,
+                          eval_,
+                          useGlobal,
+                          ignoreUndefined,
+                          replMode);
+  }
+
+  var options, input, output, dom;
+  if (prompt !== null && typeof prompt === 'object') {
     // an options object was given
     options = prompt;
     stream = options.stream || options.socket;
@@ -93,7 +192,9 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     useGlobal = options.useGlobal;
     ignoreUndefined = options.ignoreUndefined;
     prompt = options.prompt;
-  } else if (typeof prompt != 'string') {
+    dom = options.domain;
+    replMode = options.replMode;
+  } else if (typeof prompt !== 'string') {
     throw new Error('An options Object, or a prompt String are required');
   } else {
     options = {};
@@ -101,25 +202,103 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
 
   var self = this;
 
+  self._domain = dom || domain.create();
+
   self.useGlobal = !!useGlobal;
   self.ignoreUndefined = !!ignoreUndefined;
+  self.replMode = replMode || exports.REPL_MODE_SLOPPY;
 
-  self.eval = eval_ || function(code, context, file, cb) {
-    var err, result;
-    try {
-      if (self.useGlobal) {
-        result = vm.runInThisContext(code, file);
-      } else {
-        result = vm.runInContext(code, context, file);
+  self._inTemplateLiteral = false;
+
+  // just for backwards compat, see github.com/joyent/node/pull/7127
+  self.rli = this;
+
+  const savedRegExMatches = ['', '', '', '', '', '', '', '', '', ''];
+  const sep = '\u0000\u0000\u0000';
+  const regExMatcher = new RegExp(`^${sep}(.*)${sep}(.*)${sep}(.*)${sep}(.*)` +
+                                  `${sep}(.*)${sep}(.*)${sep}(.*)${sep}(.*)` +
+                                  `${sep}(.*)$`);
+
+  eval_ = eval_ || defaultEval;
+
+  function defaultEval(code, context, file, cb) {
+    var err, result, retry = false;
+    // first, create the Script object to check the syntax
+    while (true) {
+      try {
+        if (!/^\s*$/.test(code) &&
+            (self.replMode === exports.REPL_MODE_STRICT || retry)) {
+          // "void 0" keeps the repl from returning "use strict" as the
+          // result value for let/const statements.
+          code = `'use strict'; void 0;\n${code}`;
+        }
+        var script = vm.createScript(code, {
+          filename: file,
+          displayErrors: false
+        });
+      } catch (e) {
+        debug('parse error %j', code, e);
+        if (self.replMode === exports.REPL_MODE_MAGIC &&
+            e.message === BLOCK_SCOPED_ERROR &&
+            !retry) {
+          retry = true;
+          continue;
+        }
+        if (isRecoverableError(e, self))
+          err = new Recoverable(e);
+        else
+          err = e;
       }
-    } catch (e) {
-      err = e;
+      break;
     }
-    cb(err, result);
-  };
 
-  self.resetContext();
-  self.bufferedCommand = '';
+    // This will set the values from `savedRegExMatches` to corresponding
+    // predefined RegExp properties `RegExp.$1`, `RegExp.$2` ... `RegExp.$9`
+    regExMatcher.test(savedRegExMatches.join(sep));
+
+    if (!err) {
+      try {
+        if (self.useGlobal) {
+          result = script.runInThisContext({ displayErrors: false });
+        } else {
+          result = script.runInContext(context, { displayErrors: false });
+        }
+      } catch (e) {
+        err = e;
+        if (err && process.domain) {
+          debug('not recoverable, send to domain');
+          process.domain.emit('error', err);
+          process.domain.exit();
+          return;
+        }
+      }
+    }
+
+    // After executing the current expression, store the values of RegExp
+    // predefined properties back in `savedRegExMatches`
+    for (let idx = 1; idx < savedRegExMatches.length; idx += 1) {
+      savedRegExMatches[idx] = RegExp[`$${idx}`];
+    }
+
+    cb(err, result);
+  }
+
+  self.eval = self._domain.bind(eval_);
+
+  self._domain.on('error', function(e) {
+    debug('domain error');
+    const top = replMap.get(self);
+    internalUtil.decorateErrorStack(e);
+    if (e.stack && self.replMode === exports.REPL_MODE_STRICT) {
+      e.stack = e.stack.replace(/(\s+at\s+repl:)(\d+)/,
+                                (_, pre, line) => pre + (line - 1));
+    }
+    top.outputStream.write((e.stack || e) + '\n');
+    top.lineParser.reset();
+    top.bufferedCommand = '';
+    top.lines.level = [];
+    top.displayPrompt();
+  });
 
   if (!input && !output) {
     // legacy API, passing a 'stream'/'socket' option
@@ -138,32 +317,36 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     }
   }
 
-
   self.inputStream = input;
   self.outputStream = output;
 
-  self.prompt = (prompt != undefined ? prompt : '> ');
+  self.resetContext();
+  self.lineParser = new LineParser();
+  self.bufferedCommand = '';
+  self.lines.level = [];
 
   function complete(text, callback) {
     self.complete(text, callback);
   }
 
-  var rli = rl.createInterface({
+  rl.Interface.call(this, {
     input: self.inputStream,
     output: self.outputStream,
     completer: complete,
-    terminal: options.terminal
+    terminal: options.terminal,
+    historySize: options.historySize
   });
-  self.rli = rli;
 
-  this.commands = {};
+  self.setPrompt(prompt !== undefined ? prompt : '> ');
+
+  this.commands = Object.create(null);
   defineDefaultCommands(this);
 
   // figure out which "writer" function to use
   self.writer = options.writer || exports.writer;
 
-  if (typeof options.useColors === 'undefined') {
-    options.useColors = rli.terminal;
+  if (options.useColors === undefined) {
+    options.useColors = self.terminal;
   }
   self.useColors = !!options.useColors;
 
@@ -174,42 +357,51 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     };
   }
 
-  rli.setPrompt(self.prompt);
+  self.setPrompt(self._prompt);
 
-  rli.on('close', function() {
+  self.on('close', function() {
     self.emit('exit');
   });
 
   var sawSIGINT = false;
-  rli.on('SIGINT', function() {
-    var empty = rli.line.length === 0;
-    rli.clearLine();
+  self.on('SIGINT', function() {
+    var empty = self.line.length === 0;
+    self.clearLine();
 
     if (!(self.bufferedCommand && self.bufferedCommand.length > 0) && empty) {
       if (sawSIGINT) {
-        rli.close();
+        self.close();
         sawSIGINT = false;
         return;
       }
-      rli.output.write('(^C again to quit)\n');
+      self.output.write('(To exit, press ^C again or type .exit)\n');
       sawSIGINT = true;
     } else {
       sawSIGINT = false;
     }
 
+    self.lineParser.reset();
     self.bufferedCommand = '';
+    self.lines.level = [];
     self.displayPrompt();
   });
 
-  rli.on('line', function(cmd) {
+  self.on('line', function(cmd) {
+    debug('line %j', cmd);
     sawSIGINT = false;
     var skipCatchall = false;
-    cmd = trimWhitespace(cmd);
+
+    // leading whitespaces in template literals should not be trimmed.
+    if (self._inTemplateLiteral) {
+      self._inTemplateLiteral = false;
+    } else {
+      cmd = self.lineParser.parseLine(cmd);
+    }
 
     // Check to see if a REPL keyword was used. If it returns true,
     // display next prompt and return.
-    if (cmd && cmd.charAt(0) === '.') {
-      var matches = cmd.match(/^(\.[^\s]+)\s*(.*)$/);
+    if (cmd && cmd.charAt(0) === '.' && isNaN(parseFloat(cmd))) {
+      var matches = cmd.match(/^\.([^\s]+)\s*(.*)$/);
       var keyword = matches && matches[1];
       var rest = matches && matches[2];
       if (self.parseREPLKeyword(keyword, rest) === true) {
@@ -220,159 +412,177 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
       }
     }
 
-    // Check if a builtin module name was used and then include it
-    // if there's no conflict.
-    if (exports._builtinLibs.indexOf(cmd) !== -1) {
-      var lib = require(cmd);
-      if (cmd in self.context && lib !== self.context[cmd]) {
-        self.outputStream.write('A different "' + cmd +
-                                '" already exists globally\n');
+    if (!skipCatchall && (cmd || (!cmd && self.bufferedCommand))) {
+      var evalCmd = self.bufferedCommand + cmd;
+      if (/^\s*\{/.test(evalCmd) && /\}\s*$/.test(evalCmd)) {
+        // It's confusing for `{ a : 1 }` to be interpreted as a block
+        // statement rather than an object literal.  So, we first try
+        // to wrap it in parentheses, so that it will be interpreted as
+        // an expression.
+        evalCmd = '(' + evalCmd + ')\n';
       } else {
-        self.context._ = self.context[cmd] = lib;
-        self.outputStream.write(self.writer(lib) + '\n');
+        // otherwise we just append a \n so that it will be either
+        // terminated, or continued onto the next expression if it's an
+        // unexpected end of input.
+        evalCmd = evalCmd + '\n';
       }
-      self.displayPrompt();
-      return;
-    }
-
-    if (!skipCatchall) {
-      var evalCmd = self.bufferedCommand + cmd + '\n';
-
-      // This try is for determining if the command is complete, or should
-      // continue onto the next line.
-      // We try to evaluate both expressions e.g.
-      //  '{ a : 1 }'
-      // and statements e.g.
-      //  'for (var i = 0; i < 10; i++) console.log(i);'
-
-      // First we attempt to eval as expression with parens.
-      // This catches '{a : 1}' properly.
-      self.eval('(' + evalCmd + ')',
-                self.context,
-                'repl',
-                function(e, ret) {
-            if (e && !isSyntaxError(e)) return finish(e);
-
-            if (typeof ret === 'function' &&
-                /^[\r\n\s]*function/.test(evalCmd) ||
-                e) {
-              // Now as statement without parens.
-              self.eval(evalCmd, self.context, 'repl', finish);
-            } else {
-              finish(null, ret);
-            }
-          });
 
+      debug('eval %j', evalCmd);
+      self.eval(evalCmd, self.context, 'repl', finish);
     } else {
       finish(null);
     }
 
-    function isSyntaxError(e) {
-      // Convert error to string
-      e = e && (e.stack || e.toString());
-      return e && e.match(/^SyntaxError/) &&
-             !(e.match(/^SyntaxError: Unexpected token .*\n/) &&
-             e.match(/\n    at Object.parse \(native\)\n/));
-    }
-
     function finish(e, ret) {
-
+      debug('finish', e, ret);
       self.memory(cmd);
 
-      // If error was SyntaxError and not JSON.parse error
-      if (isSyntaxError(e)) {
-        if (!self.bufferedCommand && cmd.trim().match(/^npm /)) {
-          self.outputStream.write('npm should be run outside of the ' +
-                                  'node repl, in your normal shell.\n' +
-                                  '(Press Control-D to exit.)\n');
-          self.bufferedCmd = '';
-          self.displayPrompt();
-          return;
-        }
-
-        // Start buffering data like that:
-        // {
-        // ...  x: 1
-        // ... }
-        self.bufferedCommand += cmd + '\n';
+      if (e && !self.bufferedCommand && cmd.trim().startsWith('npm ')) {
+        self.outputStream.write('npm should be run outside of the ' +
+                                'node repl, in your normal shell.\n' +
+                                '(Press Control-D to exit.)\n');
+        self.lineParser.reset();
+        self.bufferedCommand = '';
         self.displayPrompt();
         return;
-      } else if (e) {
-        self.outputStream.write((e.stack || e) + '\n');
+      }
+
+      // If error was SyntaxError and not JSON.parse error
+      if (e) {
+        if (e instanceof Recoverable && !self.lineParser.shouldFail) {
+          // Start buffering data like that:
+          // {
+          // ...  x: 1
+          // ... }
+          self.bufferedCommand += cmd + '\n';
+          self.displayPrompt();
+          return;
+        } else {
+          self._domain.emit('error', e.err || e);
+        }
       }
 
       // Clear buffer if no SyntaxErrors
+      self.lineParser.reset();
       self.bufferedCommand = '';
 
       // If we got any output - print it (if no error)
-      if (!e && (!self.ignoreUndefined || ret !== undefined)) {
+      if (!e &&
+          // When an invalid REPL command is used, error message is printed
+          // immediately. We don't have to print anything else. So, only when
+          // the second argument to this function is there, print it.
+          arguments.length === 2 &&
+          (!self.ignoreUndefined || ret !== undefined)) {
         self.context._ = ret;
         self.outputStream.write(self.writer(ret) + '\n');
       }
 
       // Display prompt again
       self.displayPrompt();
-    };
+    }
   });
 
-  rli.on('SIGCONT', function() {
+  self.on('SIGCONT', function() {
     self.displayPrompt(true);
   });
 
   self.displayPrompt();
 }
-inherits(REPLServer, EventEmitter);
+inherits(REPLServer, rl.Interface);
 exports.REPLServer = REPLServer;
 
+exports.REPL_MODE_SLOPPY = Symbol('repl-sloppy');
+exports.REPL_MODE_STRICT = Symbol('repl-strict');
+exports.REPL_MODE_MAGIC = Symbol('repl-magic');
 
 // prompt is a string to print on each line for the prompt,
 // source is a stream to use for I/O, defaulting to stdin/stdout.
-exports.start = function(prompt, source, eval_, useGlobal, ignoreUndefined) {
-  var repl = new REPLServer(prompt, source, eval_, useGlobal, ignoreUndefined);
+exports.start = function(prompt,
+                         source,
+                         eval_,
+                         useGlobal,
+                         ignoreUndefined,
+                         replMode) {
+  var repl = new REPLServer(prompt,
+                            source,
+                            eval_,
+                            useGlobal,
+                            ignoreUndefined,
+                            replMode);
   if (!exports.repl) exports.repl = repl;
+  replMap.set(repl, repl);
   return repl;
 };
 
-
 REPLServer.prototype.createContext = function() {
-  if (!this.useGlobal) {
-    var context = vm.createContext();
-    for (var i in global) context[i] = global[i];
+  var context;
+  if (this.useGlobal) {
+    context = global;
   } else {
-    var context = global;
+    context = vm.createContext();
+    for (var i in global) context[i] = global[i];
+    context.console = new Console(this.outputStream);
+    context.global = context;
+    context.global.global = context;
   }
 
+  const module = new Module('<repl>');
+  module.paths = Module._resolveLookupPaths('<repl>', parentModule)[1];
+
+  const require = internalModule.makeRequireFunction.call(module);
   context.module = module;
   context.require = require;
-  context.global = context;
-  context.global.global = context;
 
   this.lines = [];
   this.lines.level = [];
 
+  // make built-in modules available directly
+  // (loaded lazily)
+  exports._builtinLibs.forEach(function(name) {
+    Object.defineProperty(context, name, {
+      get: function() {
+        var lib = require(name);
+        context._ = context[name] = lib;
+        return lib;
+      },
+      // allow the creation of other globals with this name
+      set: function(val) {
+        delete context[name];
+        context[name] = val;
+      },
+      configurable: true
+    });
+  });
+
   return context;
 };
 
-REPLServer.prototype.resetContext = function(force) {
-  if (!context || force) {
-    context = this.createContext();
-    for (var i in require.cache) delete require.cache[i];
-  }
+REPLServer.prototype.resetContext = function() {
+  this.context = this.createContext();
 
-  this.context = context;
+  // Allow REPL extensions to extend the new context
+  this.emit('reset', this.context);
 };
 
 REPLServer.prototype.displayPrompt = function(preserveCursor) {
-  var prompt = this.prompt;
+  var prompt = this._initialPrompt;
   if (this.bufferedCommand.length) {
     prompt = '...';
-    var levelInd = new Array(this.lines.level.length).join('..');
+    const len = this.lines.level.length ? this.lines.level.length - 1 : 0;
+    const levelInd = '..'.repeat(len);
     prompt += levelInd + ' ';
   }
-  this.rli.setPrompt(prompt);
-  this.rli.prompt(preserveCursor);
+
+  // Do not overwrite `_initialPrompt` here
+  REPLServer.super_.prototype.setPrompt.call(this, prompt);
+  this.prompt(preserveCursor);
 };
 
+// When invoked as an API method, overwrite _initialPrompt
+REPLServer.prototype.setPrompt = function setPrompt(prompt) {
+  this._initialPrompt = prompt;
+  REPLServer.super_.prototype.setPrompt.call(this, prompt);
+};
 
 // A stream to push an array into a REPL
 // used in REPLServer.complete
@@ -384,7 +594,7 @@ function ArrayStream() {
     data.forEach(function(line) {
       self.emit('data', line + '\n');
     });
-  }
+  };
 }
 util.inherits(ArrayStream, Stream);
 ArrayStream.prototype.readable = true;
@@ -392,10 +602,19 @@ ArrayStream.prototype.writable = true;
 ArrayStream.prototype.resume = function() {};
 ArrayStream.prototype.write = function() {};
 
-var requireRE = /\brequire\s*\(['"](([\w\.\/-]+\/)?([\w\.\/-]*))/;
-var simpleExpressionRE =
+const requireRE = /\brequire\s*\(['"](([\w\.\/-]+\/)?([\w\.\/-]*))/;
+const simpleExpressionRE =
     /(([a-zA-Z_$](?:\w|\$)*)\.)*([a-zA-Z_$](?:\w|\$)*)\.?$/;
 
+function intFilter(item) {
+  // filters out anything not starting with A-Z, a-z, $ or _
+  return /^[A-Za-z_$]/.test(item);
+}
+
+function filteredOwnPropertyNames(obj) {
+  if (!obj) return [];
+  return Object.getOwnPropertyNames(obj).filter(intFilter);
+}
 
 // Provide a list of completions for the given leading text. This is
 // given to the readline interface for handling tab completion.
@@ -409,7 +628,7 @@ var simpleExpressionRE =
 // getter code.
 REPLServer.prototype.complete = function(line, callback) {
   // There may be local variables to evaluate, try a nested REPL
-  if (this.bufferedCommand != undefined && this.bufferedCommand.length) {
+  if (this.bufferedCommand !== undefined && this.bufferedCommand.length) {
     // Get a new array of inputed lines
     var tmp = this.lines.slice();
     // Kill off all function declarations to push all local variables into
@@ -426,6 +645,7 @@ REPLServer.prototype.complete = function(line, callback) {
     // all this is only profitable if the nested REPL
     // does not have a bufferedCommand
     if (!magic.bufferedCommand) {
+      replMap.set(magic, replMap.get(this));
       return magic.complete(line, callback);
     }
   }
@@ -435,7 +655,7 @@ REPLServer.prototype.complete = function(line, callback) {
   // list of completion lists, one for each inheritance "level"
   var completionGroups = [];
 
-  var completeOn, match, filter, i, j, group, c;
+  var completeOn, i, group, c;
 
   // REPL commands (e.g. ".break").
   var match = null;
@@ -450,11 +670,7 @@ REPLServer.prototype.complete = function(line, callback) {
     completionGroupsLoaded();
   } else if (match = line.match(requireRE)) {
     // require('...<Tab>')
-    //TODO: suggest require.exts be exposed to be introspec registered
-    //extensions?
-    //TODO: suggest include the '.' in exts in internal repr: parity with
-    //`path.extname`.
-    var exts = ['.js', '.node'];
+    const exts = Object.keys(this.context.require.extensions);
     var indexRe = new RegExp('^index(' + exts.map(regexpEscape).join('|') +
                              ')$');
 
@@ -537,23 +753,21 @@ REPLServer.prototype.complete = function(line, callback) {
       }
 
       // Resolve expr and get its completions.
-      var obj, memberGroups = [];
+      var memberGroups = [];
       if (!expr) {
         // If context is instance of vm.ScriptContext
         // Get global vars synchronously
-        if (this.useGlobal ||
-            this.context.constructor &&
-            this.context.constructor.name === 'Context') {
+        if (this.useGlobal || vm.isContext(this.context)) {
           var contextProto = this.context;
           while (contextProto = Object.getPrototypeOf(contextProto)) {
-            completionGroups.push(Object.getOwnPropertyNames(contextProto));
+            completionGroups.push(filteredOwnPropertyNames(contextProto));
           }
-          completionGroups.push(Object.getOwnPropertyNames(this.context));
+          completionGroups.push(filteredOwnPropertyNames(this.context));
           addStandardGlobals(completionGroups, filter);
           completionGroupsLoaded();
         } else {
           this.eval('.scope', this.context, 'repl', function(err, globals) {
-            if (err || !globals) {
+            if (err || !Array.isArray(globals)) {
               addStandardGlobals(completionGroups, filter);
             } else if (Array.isArray(globals[0])) {
               // Add grouped globals
@@ -573,7 +787,14 @@ REPLServer.prototype.complete = function(line, callback) {
 
           if (obj != null) {
             if (typeof obj === 'object' || typeof obj === 'function') {
-              memberGroups.push(Object.getOwnPropertyNames(obj));
+              try {
+                memberGroups.push(filteredOwnPropertyNames(obj));
+              } catch (ex) {
+                // Probably a Proxy object without `getOwnPropertyNames` trap.
+                // We simply ignore it here, as we don't want to break the
+                // autocompletion. Fixes the bug
+                // https://github.com/nodejs/node/issues/2119
+              }
             }
             // works for non-objects
             try {
@@ -585,7 +806,7 @@ REPLServer.prototype.complete = function(line, callback) {
                 p = obj.constructor ? obj.constructor.prototype : null;
               }
               while (p !== null) {
-                memberGroups.push(Object.getOwnPropertyNames(p));
+                memberGroups.push(filteredOwnPropertyNames(p));
                 p = Object.getPrototypeOf(p);
                 // Circular refs possible? Let's guard against that.
                 sentinel--;
@@ -649,7 +870,7 @@ REPLServer.prototype.complete = function(line, callback) {
         group.sort();
         for (var j = 0; j < group.length; j++) {
           c = group[j];
-          if (!hasOwnProperty(c)) {
+          if (!hasOwnProperty(uniq, c)) {
             completions.push(c);
             uniq[c] = true;
           }
@@ -688,7 +909,7 @@ REPLServer.prototype.defineCommand = function(keyword, cmd) {
   } else if (typeof cmd.action !== 'function') {
     throw new Error('bad argument, action must be a function');
   }
-  this.commands['.' + keyword] = cmd;
+  this.commands[keyword] = cmd;
 };
 
 REPLServer.prototype.memory = function memory(cmd) {
@@ -700,7 +921,8 @@ REPLServer.prototype.memory = function memory(cmd) {
   // save the line so I can do magic later
   if (cmd) {
     // TODO should I tab the level?
-    self.lines.push(new Array(self.lines.level.length).join('  ') + cmd);
+    const len = self.lines.level.length ? self.lines.level.length - 1 : 0;
+    self.lines.push('  '.repeat(len) + cmd);
   } else {
     // I don't want to not change the format too much...
     self.lines.push('');
@@ -787,10 +1009,10 @@ function addStandardGlobals(completionGroups, filter) {
 }
 
 function defineDefaultCommands(repl) {
-  // TODO remove me after 0.3.x
   repl.defineCommand('break', {
     help: 'Sometimes you get stuck, this gets you out',
     action: function() {
+      this.lineParser.reset();
       this.bufferedCommand = '';
       this.displayPrompt();
     }
@@ -805,10 +1027,11 @@ function defineDefaultCommands(repl) {
   repl.defineCommand('clear', {
     help: clearMessage,
     action: function() {
+      this.lineParser.reset();
       this.bufferedCommand = '';
       if (!this.useGlobal) {
         this.outputStream.write('Clearing context...\n');
-        this.resetContext(true);
+        this.resetContext();
       }
       this.displayPrompt();
     }
@@ -817,7 +1040,7 @@ function defineDefaultCommands(repl) {
   repl.defineCommand('exit', {
     help: 'Exit the repl',
     action: function() {
-      this.rli.close();
+      this.close();
     }
   });
 
@@ -858,9 +1081,12 @@ function defineDefaultCommands(repl) {
           this.displayPrompt();
           lines.forEach(function(line) {
             if (line) {
-              self.rli.write(line + '\n');
+              self.write(line + '\n');
             }
           });
+        } else {
+          this.outputStream.write('Failed to load:' + file +
+                                  ' is not a valid file\n');
         }
       } catch (e) {
         this.outputStream.write('Failed to load:' + file + '\n');
@@ -870,18 +1096,6 @@ function defineDefaultCommands(repl) {
   });
 }
 
-
-function trimWhitespace(cmd) {
-  var trimmer = /^\s*(.+)\s*$/m,
-      matches = trimmer.exec(cmd);
-
-  if (matches && matches.length === 2) {
-    return matches[1];
-  }
-  return '';
-}
-
-
 function regexpEscape(s) {
   return s.replace(/[-[\]{}()*+?.,\\^$|#\s]/g, '\\$&');
 }
@@ -896,9 +1110,9 @@ function regexpEscape(s) {
  * @return {String} The converted command.
  */
 REPLServer.prototype.convertToContext = function(cmd) {
-  var self = this, matches,
-      scopeVar = /^\s*var\s*([_\w\$]+)(.*)$/m,
-      scopeFunc = /^\s*function\s*([_\w\$]+)/;
+  const scopeVar = /^\s*var\s*([_\w\$]+)(.*)$/m;
+  const scopeFunc = /^\s*function\s*([_\w\$]+)/;
+  var matches;
 
   // Replaces: var foo = "bar";  with: self.context.foo = bar;
   matches = scopeVar.exec(cmd);
@@ -907,10 +1121,31 @@ REPLServer.prototype.convertToContext = function(cmd) {
   }
 
   // Replaces: function foo() {};  with: foo = function foo() {};
-  matches = scopeFunc.exec(self.bufferedCommand);
+  matches = scopeFunc.exec(this.bufferedCommand);
   if (matches && matches.length === 2) {
-    return matches[1] + ' = ' + self.bufferedCommand;
+    return matches[1] + ' = ' + this.bufferedCommand;
   }
 
   return cmd;
 };
+
+
+// If the error is that we've unexpectedly ended the input,
+// then let the user try to recover by adding more input.
+function isRecoverableError(e, self) {
+  if (e && e.name === 'SyntaxError') {
+    var message = e.message;
+    if (message === 'Unterminated template literal' ||
+        message === 'Missing } in template expression') {
+      self._inTemplateLiteral = true;
+      return true;
+    }
+    return /^(Unexpected end of input|Unexpected token)/.test(message);
+  }
+  return false;
+}
+
+function Recoverable(err) {
+  this.err = err;
+}
+inherits(Recoverable, SyntaxError);
