@@ -17,17 +17,25 @@ module.exports = function(context) {
     return {
 
         "Program": function checkBadEOF(node) {
-            // Get the whole source code, not for node only.
-            var src = context.getSource(), location = {column: 1};
 
-            if (src.length === 0) {
-                return;
-            }
+            // Get the whole source code, not for node only.
+            var src = context.getSource(),
+                location = {column: 1},
+                linebreakStyle = context.options[0] || "unix",
+                linebreak = linebreakStyle === "unix" ? "\n" : "\r\n";
 
             if (src[src.length - 1] !== "\n") {
+
                 // file is not newline-terminated
                 location.line = src.split(/\n/g).length;
-                context.report(node, location, "Newline required at end of file but not found.");
+                context.report({
+                    node: node,
+                    loc: location,
+                    message: "Newline required at end of file but not found.",
+                    fix: function(fixer) {
+                        return fixer.insertTextAfterRange([0, src.length], linebreak);
+                    }
+                });
             }
         }
 
@@ -35,4 +43,8 @@ module.exports = function(context) {
 
 };
 
-module.exports.schema = [];
+module.exports.schema = [
+    {
+        "enum": ["unix", "windows"]
+    }
+];
